// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {
    crate::api_client::ApiClient,
    crate::commands::import,
    crate::constants,
    crate::recipe,
    anyhow::{bail, Result},
    chrono::naive::{NaiveDate, NaiveDateTime, NaiveTime},
    chrono::{DateTime, Local, TimeZone},
    futures::future::try_join_all,
    ics::properties as calprop,
    ics::Event,
    reqwest::StatusCode,
    std::collections::{HashMap, HashSet},
    std::fmt::Write,
    std::iter::Iterator,
    std::path::Path,
};

#[derive(serde::Deserialize)]
struct CsvRecord {
    day: NaiveDate,
    lunch: String,
    dinner: String,
}

#[derive(strum_macros::Display)]
enum Meal {
    Lunch,
    Dinner,
}

pub async fn with(api_client: &ApiClient, calendar: &str, csv_file: &Path) -> Result<()> {
    let mut csv = csv::Reader::from_path(csv_file)?;
    let records = csv.deserialize::<CsvRecord>().flatten().collect::<Vec<_>>();

    let recipe_urls = urls_from_csv(records.iter())?;
    import::with(&api_client, recipe_urls.into_iter()).await?;
    let recipes = get_all_recipes(&api_client).await?;

    let events = records
        .iter()
        .flat_map(|r| {
            let lunch = recipes.get(&r.lunch);
            let dinner = recipes.get(&r.dinner);

            let events = [
                lunch.map(|recipe| to_event(r.day, Meal::Lunch, recipe)),
                dinner.map(|recipe| to_event(r.day, Meal::Dinner, recipe)),
            ];

            events
        })
        .flatten();

    publish_events(&api_client, calendar, events).await?;
    Ok(())
}

fn to_event(date: NaiveDate, meal: Meal, recipe: &recipe::Recipe) -> (String, Event) {
    // TODO: this is momentarily hardcoded, should be an option
    let meal_time = match meal {
        Meal::Lunch => NaiveTime::from_hms(12, 00, 00),
        Meal::Dinner => NaiveTime::from_hms(19, 00, 00),
    };

    let uid = format!(
        "{}-{}@{}.montecristosoftware.eu",
        date,
        meal,
        env!("CARGO_PKG_NAME")
    );

    let end_time = NaiveDateTime::new(date, meal_time);
    let start_time = end_time - recipe.total_time();

    let mut event = Event::new(uid.clone(), dt_fmt(&Local::now()));
    event.push(calprop::Summary::new(&recipe.name));
    event.push(calprop::Description::new(format!("cookbook@{}", recipe.id)));
    event.push(calprop::Location::new(&recipe.url));
    event.push(calprop::DtStart::new(dt_fmt(
        &Local.from_local_datetime(&start_time).unwrap(),
    )));
    event.push(calprop::DtEnd::new(dt_fmt(
        &Local.from_local_datetime(&end_time).unwrap(),
    )));

    // TODO make configurable yearly repetition, for now this suits my personal uses
    event.push(calprop::RRule::new("FREQ=YEARLY"));

    let alarm = ics::Alarm::display(
        calprop::Trigger::new("-P15M"),
        calprop::Description::new(&recipe.name),
    );
    event.add_alarm(alarm);

    (uid, event)
}

fn urls_from_csv<'a, RecordsIter>(records: RecordsIter) -> Result<HashSet<String>>
where
    RecordsIter: Iterator<Item = &'a CsvRecord>,
{
    Ok(
        records.fold(std::collections::HashSet::new(), |mut set, r| {
            if !r.lunch.is_empty() {
                set.insert(r.lunch.clone());
            }
            if !r.dinner.is_empty() {
                set.insert(r.dinner.clone());
            }
            set
        }),
    )
}

fn dt_fmt(datetime: &DateTime<Local>) -> String {
    datetime.format("%Y%m%dT%H%M%S").to_string()
}

async fn get_all_recipes(api_client: &ApiClient) -> Result<HashMap<String, recipe::Recipe>> {
    let metadata = api_client
        .rest
        .get(api_client.base_url.join("apps/cookbook/api/recipes")?)
        .send()
        .await?
        .json::<Vec<recipe::Metadata>>()
        .await?;

    let recipes = metadata.iter().map(|rm| async {
        let response = api_client
            .rest
            .get(
                api_client
                    .base_url
                    .join(&format!("apps/cookbook/api/recipes/{id}", id = rm.id))
                    .unwrap(),
            )
            .send()
            .await
            .expect(&format!(
                "Cannot fetch recipe {} with id {}",
                rm.name, rm.id
            ));
        response.json::<recipe::Recipe>().await
    });

    let recipes = try_join_all(recipes).await?;
    Ok(HashMap::from_iter(
        recipes.into_iter().map(|r| (r.url.clone(), r)),
    ))
}

async fn publish_events<'a, EventsIter>(
    api_client: &ApiClient,
    calendar: &str,
    events: EventsIter,
) -> Result<()>
where
    EventsIter: Iterator<Item = (String, Event<'a>)>,
{
    let calendar_prototype: ics::ICalendar = ics::ICalendar::new(
        "2.0",
        format!(
            "-//{}//NONSGML {}//EN",
            constants::VENDOR,
            env!("CARGO_PKG_NAME")
        ),
    );

    let dav_base = api_client
        .rest
        .head(api_client.base_url.join("/.well-known/caldav")?)
        .send()
        .await?;

    let calendar_url = dav_base.url().join(&format!(
        "calendars/{}/{}/",
        &api_client.username,
        calendar.to_lowercase().as_str().replace(" ", "-")
    ))?;

    let calendar_prototype = &calendar_prototype;
    let calendar_url = &calendar_url;
    let update_requests = events.map(|(event_id, event)| async move {
        let mut cal = calendar_prototype.clone();
        cal.add_event(event);

        let url = calendar_url.join(&format!("{}.ics", event_id)).unwrap();
        api_client
            .rest
            .put(url)
            .header("Content-Type", "text/calendar; charset=utf-8")
            .body(cal.to_string())
            .send()
            .await
    });

    let responses = try_join_all(update_requests).await?;
    let failed_responses = responses.into_iter().filter(|response| {
        ![StatusCode::NO_CONTENT, StatusCode::CREATED].contains(&response.status())
    });

    let mut errors = String::new();
    for r in failed_responses {
        write!(errors, "\n{}", r.text().await.unwrap())?;
    }

    if !errors.is_empty() {
        bail!("Error while updating calendar events: {}", errors);
    }

    Ok(())
}
